import reboot.aio.reboot
import unittest
from reboot.aio.applications import Application
from reboot.aio.auth.token_verifiers import TokenVerifier
from reboot.aio.contexts import EffectValidation
from reboot.aio.http import WebFramework
from reboot.aio.reboot import ApplicationRevision
from reboot.aio.servicers import Servicer
from typing import Any, Optional, overload
from unittest import mock


def assert_called_twice_with(
    testcase: unittest.IsolatedAsyncioTestCase,
    mock_obj: mock.Mock,
    *args: Any,
    **kwargs: Any,
) -> None:
    """Asserts that the given mock received exactly two calls, and that they had
    the given arguments.

    This is a useful alternative to `Mock.assert_called_once_with` in the presence
    of effect validation, which will frequently trigger two identical calls to a mock.
    """
    call = mock.call(*args, **kwargs)
    mock_obj.assert_has_calls([call, call])
    testcase.assertEqual(mock_obj.call_count, 2)


class Reboot(reboot.aio.reboot.Reboot):
    """A testing specific version of `Reboot` that takes an `Application`
    instead of explicit keyword args."""

    @overload
    async def up(
        self,
        application: Application,
        *,
        local_envoy: bool = False,
        local_envoy_port: int = 0,
        partitions: Optional[int] = None,
        effect_validation: Optional[EffectValidation] = None,
        in_process: bool = False,
    ) -> ApplicationRevision:
        ...

    @overload
    async def up(
        self,
        *,
        revision: Optional[ApplicationRevision] = None,
    ) -> ApplicationRevision:
        ...

    @overload
    async def up(
        self,
        *,
        servicers: list[type[Servicer]],
        # A legacy gRPC servicer type can't be more specific than `type`,
        # because legacy gRPC servicers (as generated by the gRPC `protoc`
        # plugin) do not share any common base class other than `object`.
        legacy_grpc_servicers: Optional[list[type]] = None,
        web_framework: WebFramework,
        token_verifier: Optional[TokenVerifier] = None,
        local_envoy: bool = False,
        local_envoy_port: int = 0,
        partitions: Optional[int] = None,
        effect_validation: Optional[EffectValidation] = None,
        in_process: bool = False,
        revision: Optional[ApplicationRevision] = None,
    ) -> ApplicationRevision:
        ...

    async def up(
        self,
        application: Optional[Application] = None,
        *,
        servicers: Optional[list[type[Servicer]]] = None,
        # A legacy gRPC servicer type can't be more specific than `type`,
        # because legacy gRPC servicers (as generated by the gRPC `protoc`
        # plugin) do not share any common base class other than `object`.
        legacy_grpc_servicers: Optional[list[type]] = None,
        web_framework: Optional[WebFramework] = None,
        token_verifier: Optional[TokenVerifier] = None,
        local_envoy: bool = False,
        local_envoy_port: int = 0,
        partitions: Optional[int] = None,
        effect_validation: Optional[EffectValidation] = None,
        in_process: bool = False,
        revision: Optional[ApplicationRevision] = None,
    ) -> ApplicationRevision:

        if revision is not None and application is not None:
            raise ValueError(
                "Only one of 'application' OR 'revision' can be passed"
            )

        if servicers is not None:
            raise ValueError("Not expecting 'servicers'")

        if legacy_grpc_servicers is not None:
            raise ValueError("Not expecting 'legacy_grpc_servicers'")

        if web_framework is not None:
            raise ValueError("Not expecting 'web_framework'")

        if token_verifier is not None:
            raise ValueError("Not expecting 'token_verifier'")

        if revision is not None:
            if partitions is not None:
                raise ValueError("Not expecting 'partitions'")

            if effect_validation is not None:
                raise ValueError("Not expecting 'effect_validation'")

            return await super().up(
                revision=revision,
            )

        if application is None:
            raise ValueError("Must pass one of 'application' or 'revision'")

        # Should only have `application`, `local_envoy`,
        # `local_envoy_port`, `partitions`, `effect_validation`, or
        # `in_process`.

        return await super().up(
            servicers=application.servicers,
            legacy_grpc_servicers=application.legacy_grpc_servicers,
            web_framework=application.web_framework,
            token_verifier=application.token_verifier,
            local_envoy=local_envoy,
            local_envoy_port=local_envoy_port,
            partitions=partitions,
            effect_validation=effect_validation,
            in_process=in_process,
        )
